#include "ScriptPositionsOverlayMode.h"
#include "OFS_ScriptTimeline.h"
#include "OFS_Profiling.h"
#include "FunscriptHeatmap.h"

#include <cmath>

ImGradient BaseOverlay::speedGradient;
std::vector<BaseOverlay::ColoredLine> BaseOverlay::ColoredLines;
std::vector<ImVec2> BaseOverlay::SelectedActionScreenCoordinates;
std::vector<ImVec2> BaseOverlay::ActionScreenCoordinates;
std::vector<FunscriptAction> BaseOverlay::ActionPositionWindow;
float BaseOverlay::PointSize = 7.f;
bool BaseOverlay::SplineMode = true;
bool BaseOverlay::ShowActions = true;
bool BaseOverlay::SyncLineEnable = false;

BaseOverlay::BaseOverlay(ScriptTimeline* timeline) noexcept
{
    OFS_PROFILE(__FUNCTION__);
    this->timeline = timeline;

    if (speedGradient.getMarks().size() == 0) {
        std::array<ImColor, 4> heatColor{
        IM_COL32(0xFF, 0xFF, 0xFF, 0xFF),
        IM_COL32(0x66, 0xff, 0x00, 0xFF),
        IM_COL32(0xFF, 0xff, 0x00, 0xFF),
        IM_COL32(0xFF, 0x00, 0x00, 0xFF),
        };

        float pos = 0.0f;
        for (auto& col : heatColor) {
            speedGradient.addMark(pos, col);
            pos += (1.f / (heatColor.size() - 1));
        }
        speedGradient.refreshCache();
    }
}

void BaseOverlay::update() noexcept
{
    OFS_PROFILE(__FUNCTION__);
    ActionScreenCoordinates.clear();
    ActionPositionWindow.clear();
    SelectedActionScreenCoordinates.clear();
}

void BaseOverlay::DrawSettings() noexcept
{

}

void EmptyOverlay::DrawScriptPositionContent(const OverlayDrawingCtx& ctx) noexcept
{
    OFS_PROFILE(__FUNCTION__);
    timeline->DrawAudioWaveform(ctx);
    BaseOverlay::DrawActionLines(ctx);
}

float EmptyOverlay::steppingIntervalForward(float fromTime) noexcept
{
    return timeline->frameTime;
}

float EmptyOverlay::steppingIntervalBackward(float fromTime) noexcept
{
    return -timeline->frameTime;
}

void BaseOverlay::DrawActionLines(const OverlayDrawingCtx& ctx) noexcept
{
    if (!BaseOverlay::ShowActions) return;
    OFS_PROFILE(__FUNCTION__);
    auto& script = *ctx.script;

    auto startIt = script.Actions().begin() + ctx.actionFromIdx;
    auto endIt = script.Actions().begin() + ctx.actionToIdx;
    ColoredLines.clear();

    auto getPointForAction = [](const OverlayDrawingCtx& ctx, FunscriptAction action) {
        float relative_x = (float)(action.atS - ctx.offsetTime) / ctx.visibleTime;
        float x = (ctx.canvas_size.x) * relative_x;
        float y = (ctx.canvas_size.y) * (1 - (action.pos / 100.f));
        x += ctx.canvas_pos.x;
        y += ctx.canvas_pos.y;
        return ImVec2(x, y);
    };

    auto drawSpline = [getPointForAction](const OverlayDrawingCtx& ctx, FunscriptAction startAction, FunscriptAction endAction, uint32_t color, float width, bool background = true)
    {
        constexpr float SamplesPerTwothousandPixels = 150.f;
        const float MaximumSamples = SamplesPerTwothousandPixels * (ctx.canvas_size.x / 2000.f);

        auto getPointForTimePos = [](const OverlayDrawingCtx& ctx, float time, float pos) noexcept {
            float relative_x = (float)(time - ctx.offsetTime) / ctx.visibleTime;
            float x = (ctx.canvas_size.x) * relative_x;
            float y = (ctx.canvas_size.y) * (1 - (pos / 100.f));
            x += ctx.canvas_pos.x;
            y += ctx.canvas_pos.y;
            return ImVec2(x, y);
        };
        auto putPoint = [getPointForTimePos](auto& ctx, float time) noexcept {
            float pos = Util::Clamp<float>(ctx.script->Spline(time) * 100.f, 0.f, 100.f);
            ctx.draw_list->PathLineTo(getPointForTimePos(ctx, time, pos));
        };

        ctx.draw_list->PathClear();
        float visibleDuration;
        float currentTime;
        float endTime;
        if (startAction.atS >= ctx.offsetTime && endAction.atS <= (ctx.offsetTime + ctx.visibleTime)) {
            currentTime = startAction.atS;
            endTime = endAction.atS;
            visibleDuration = endTime - currentTime;
        }
        else if (startAction.atS < ctx.offsetTime && endAction.atS > (ctx.offsetTime + ctx.visibleTime)) {
            // clip at the invisible area in both direction
            currentTime = ctx.offsetTime;
            endTime = (ctx.offsetTime + ctx.visibleTime);
            visibleDuration = ctx.visibleTime;
        }
        else if (startAction.atS < ctx.offsetTime) {
            // clip invisible area on the left
            currentTime = ctx.offsetTime;
            endTime = endAction.atS;
            visibleDuration = endAction.atS - ctx.offsetTime;
        }
        else if (endAction.atS > (ctx.offsetTime + ctx.visibleTime)) {
            // clip invisble area on the right
            currentTime = startAction.atS;
            endTime = (ctx.offsetTime + ctx.visibleTime) + 0.001f;
            visibleDuration = endTime - startAction.atS;
        }

        // detail gets dynamically reduced by increasing the timeStep,
        // at which is being sampled from the spline
        const float ratio = visibleDuration / ctx.visibleTime;
        const float SampleCount = MaximumSamples * ratio;

        const float timeStep = visibleDuration / SampleCount;
        
        if (SampleCount < 3.f) {
            auto p1 = getPointForAction(ctx, startAction);
            auto p2 = getPointForAction(ctx, endAction);
            if (background) {
                ctx.draw_list->PathLineTo(p1);
                ctx.draw_list->PathLineTo(p2);
                ctx.draw_list->PathStroke(IM_COL32_BLACK, false, 7.f);
            }
            ColoredLines.emplace_back(std::move(BaseOverlay::ColoredLine{ p1, p2, color }));
        }
        else {
            putPoint(ctx, currentTime);
            currentTime += timeStep;
            while (currentTime < endTime) {
                putPoint(ctx, currentTime);
                currentTime += timeStep;
            }
            putPoint(ctx, endAction.atS);
            auto tmpSize = ctx.draw_list->_Path.Size;
            ctx.draw_list->PathStroke(IM_COL32_BLACK, false, 7.f);
            ctx.draw_list->_Path.Size = tmpSize;
            ctx.draw_list->PathStroke(color, false, width);
        }
    };

    if (SplineMode) {
        const FunscriptAction* prevAction = nullptr;
        for (; startIt != endIt; startIt++) {
            auto& action = *startIt;
            auto p1 = getPointForAction(ctx, action);
            ActionScreenCoordinates.emplace_back(p1);
            ActionPositionWindow.emplace_back(action);

            if (prevAction != nullptr) {
                // calculate speed relative to maximum speed
                float rel_speed = Util::Clamp<float>((std::abs(action.pos - prevAction->pos) / ((action.atS - prevAction->atS))) / HeatmapGradient::MaxSpeedPerSecond, 0.f, 1.f);
                ImColor speed_color;
                speedGradient.getColorAt(rel_speed, &speed_color.Value.x);
                speed_color.Value.w = 1.f;

                drawSpline(ctx, *prevAction, action, ImGui::ColorConvertFloat4ToU32(speed_color), 3.f);
            }
            prevAction = &action;
        }
    }
    else {
        const FunscriptAction* prevAction = nullptr;
        for (; startIt != endIt; startIt++) {
            auto& action = *startIt;

            auto p1 = getPointForAction(ctx, action);
            ActionScreenCoordinates.emplace_back(p1);
            ActionPositionWindow.emplace_back(action);

            if (prevAction != nullptr) {
                // draw line
                auto p2 = getPointForAction(ctx, *prevAction);
                // calculate speed relative to maximum speed
                float rel_speed = Util::Clamp<float>((std::abs(action.pos - prevAction->pos) / ((action.atS - prevAction->atS))) / HeatmapGradient::MaxSpeedPerSecond, 0.f, 1.f);
                ImColor speed_color;
                speedGradient.getColorAt(rel_speed, &speed_color.Value.x);
                speed_color.Value.w = 1.f;
                
                ctx.draw_list->AddLine(p1, p2, IM_COL32(0, 0, 0, 255), 7.0f); // border
                ColoredLines.emplace_back(std::move(BaseOverlay::ColoredLine{ p1, p2, ImGui::ColorConvertFloat4ToU32(speed_color) }));
            }

            prevAction = &action;
        }

    }

    if (script.HasSelection()) {
        auto startIt = std::find_if(script.Selection().begin(), script.Selection().end(),
            [&](auto act) { return act.atS >= ctx.offsetTime; });
        if (startIt != script.Selection().begin())
            startIt -= 1;

        auto endIt = std::find_if(startIt, script.Selection().end(),
            [&](auto act) { return act.atS >= ctx.offsetTime + ctx.visibleTime; });
        if (endIt != script.Selection().end())
            endIt += 1;

        constexpr auto selectedLines = IM_COL32(3, 194, 252, 255);
        if (SplineMode) {
            const FunscriptAction* prev_action = nullptr;
            for (; startIt != endIt; startIt++) {
                auto&& action = *startIt;
                auto point = getPointForAction(ctx, action);

                if (prev_action != nullptr) {
                    // draw highlight line
                    drawSpline(ctx, *prev_action, action, selectedLines, 3.f, false);
                }

                SelectedActionScreenCoordinates.emplace_back(point);
                prev_action = &action;
            }
        }
        else {
            const FunscriptAction* prev_action = nullptr;
            for (; startIt != endIt; startIt++) {
                auto&& action = *startIt;
                auto point = getPointForAction(ctx, action);

                if (prev_action != nullptr) {
                    // draw highlight line
                    //ctx.draw_list->AddLine(getPointForAction(ctx, *prev_action), point, selectedLines, 3.0f);
                    ColoredLines.emplace_back(std::move(BaseOverlay::ColoredLine{ getPointForAction(ctx, *prev_action), point, selectedLines }));
                }

                SelectedActionScreenCoordinates.emplace_back(point);
                prev_action = &action;
            }
        }
    }

    // this is so that the black background line gets rendered first
    for (auto&& line : ColoredLines) {
        ctx.draw_list->AddLine(line.p1, line.p2, line.color, 3.f);
    }
}

void BaseOverlay::DrawSecondsLabel(const OverlayDrawingCtx& ctx) noexcept
{
    auto& style = ImGui::GetStyle();
    if (ctx.scriptIdx == ctx.drawnScriptCount - 1) {
        OFS_PROFILE(__FUNCTION__);
        char tmp[16];
        stbsp_snprintf(tmp, sizeof(tmp), "%.2f seconds", ctx.visibleTime);
        auto textSize = ImGui::CalcTextSize(tmp);
        ctx.draw_list->AddText(
            ctx.canvas_pos + ImVec2(style.FramePadding.x, ctx.canvas_size.y - textSize.y - style.FramePadding.y),
            ImGui::ColorConvertFloat4ToU32(style.Colors[ImGuiCol_Text]),
            tmp
        );
    }
}

void BaseOverlay::DrawHeightLines(const OverlayDrawingCtx& ctx) noexcept
{
    OFS_PROFILE(__FUNCTION__);
    // height indicators
    for (int i = 0; i < 9; i++) {
        auto color = (i == 4) ? IM_COL32(150, 150, 150, 255) : IM_COL32(80, 80, 80, 255);
        auto thickness = (i == 4) ? 2.f : 1.0f;
        ctx.draw_list->AddLine(
            ctx.canvas_pos + ImVec2(0.0, (ctx.canvas_size.y / 10.f) * (i + 1)),
            ctx.canvas_pos + ImVec2(ctx.canvas_size.x, (ctx.canvas_size.y / 10.f) * (i + 1)),
            color,
            thickness
        );
    }
}

void BaseOverlay::DrawScriptLabel(const OverlayDrawingCtx& ctx) noexcept
{
    OFS_PROFILE(__FUNCTION__);
    auto& style = ImGui::GetStyle();
    auto& title = ctx.script->Title;
    auto textSize = ImGui::CalcTextSize(title.c_str());
    ctx.draw_list->AddText(
        ctx.canvas_pos + ctx.canvas_size - style.FramePadding - textSize,
        ImGui::ColorConvertFloat4ToU32(style.Colors[ImGuiCol_Text]),
        title.c_str()
    );
}
